
local Mood_dst = Class(function(self, inst)
    self.inst = inst

    self.enabled = true

    self.moodtimeindays = {length = nil, wait = nil}
    self.isinmood = false
    self.daystomoodchange = nil
    self.onentermood = nil
    self.onleavemood = nil
    self.moodseasons = {}
    self.firstseasonadded = false

    inst:ListenForEvent("daycomplete", function(inst, data)
        if self.daystomoodchange and self.daystomoodchange > 0 then
            self.daystomoodchange = self.daystomoodchange - 1
            self:CheckForMoodChange()
        end
    end, GetWorld())
end)

function Mood_dst:GetDebugString()
    return string.format("inmood:%s, days till change:%s %s", self.enabled and tostring(self.isinmood) or "DISABLED", tostring(self.daystomoodchange), self.seasonmood and "SEASONMOOD" or "" )
end

function Mood_dst:Enable(enabled)
    self.enabled = enabled
    self:SetIsInMood(false, false)
end

function Mood_dst:SetMoodTimeInDays(length, wait)
    self.moodtimeindays.length = length
    self.moodtimeindays.wait = wait
    self.daystomoodchange = wait
    self.isinmood = false
end

local function OnSeasonChange(inst, season)
    if not inst.components.mood_dst or not inst.components.mood_dst.enabled then
        return
    end

	local active = false
	if inst.components.mood_dst.moodseasons then 
	    for i, s in pairs(inst.components.mood_dst.moodseasons) do
	        if s == season then
	            active = true
	            break
	        end
	    end
	end
    if active then
        inst.components.mood_dst:SetIsInMood(true, true)
    else
        inst.components.mood_dst:ResetMood()
    end        
end

function Mood_dst:ValidateMood()
	OnSeasonChange(self.inst,GetSeasonManager():GetSeason())
end

function Mood_dst:SetMoodSeason(activeseason)
    table.insert(self.moodseasons, activeseason)
    if not self.firstseasonadded then
           self.inst:ListenForEvent("seasonChange", function(it, data)
			if not self.enabled then
				return
			end
                local active = false
                for i, s in pairs(self.moodseasons) do
                    if s == data.season then
                        active = true
                        break
                    end
                end
                if active then
                    self:SetIsInMood(true, true)
                else
                    self:ResetMood()
                end        
            end, GetWorld())
        self.firstseasonadded = true
    end
end

function Mood_dst:CheckForMoodChange()
    if self.daystomoodchange == 0 then
        self:SetIsInMood(not self:IsInMood() )
    end
end

function Mood_dst:SetInMoodFn(fn)
    self.onentermood = fn
end

function Mood_dst:SetLeaveMoodFn(fn)
    self.onleavemood = fn
end

function Mood_dst:ResetMood()
    if self.seasonmood then
        self.seasonmood = false
        self.isinmood = false
        self.daystomoodchange = self.moodtimeindays.wait
        if self.onleavemood then
            self.onleavemood(self.inst)
        end
    end
end

local function GetSeasonLength()
	return GetSeasonManager() and GetSeasonManager():GetSeasonLength()
end

function Mood_dst:SetIsInMood(inmood, entireseason)
    if inmood and (not self.enabled or self.moodtimeindays.length == 0) then
        return
    end

    if self.isinmood ~= inmood or entireseason then
    
        self.isinmood = inmood
        if self.isinmood then
            if entireseason then
                self.seasonmood = true
                self.daystomoodchange = GetSeasonLength() or self.moodtimeindays.length
            else
                self.seasonmood = false
                self.daystomoodchange = self.moodtimeindays.length
            end
            if self.onentermood then
                self.onentermood(self.inst)
            end
        else
            if not entireseason then
                self.seasonmood = false
                self.daystomoodchange = self.moodtimeindays.wait
            end
            if self.onleavemood then
                self.onleavemood(self.inst)
            end
        end
    end
end

function Mood_dst:IsInMood()
    return self.isinmood
end

function Mood_dst:OnSave()
    return {inmood = self.isinmood, daysleft = self.daystomoodchange, moodseasons = self.moodseasons }
end

function Mood_dst:OnLoad(data)
	self.moodseasons = data.moodseasons or self.moodseasons
    self.isinmood = not data.inmood
    local active = false
	local season = GetSeasonManager() and GetSeasonManager():GetSeason()
    if self.moodseasons then 
	    for i, s in pairs(self.moodseasons) do
	        if season and s == season then
	            active = true
	            break
	        end
	    end
	end
    self:SetIsInMood(data.inmood, active)
    self.daystomoodchange = data.daysleft
end

return Mood_dst
