local prefabs =
{
    "babybeefalo",
}

local function InMood(inst)
    if inst.components.periodicspawner ~= nil then
        inst.components.periodicspawner:Start()
    end
    if inst.components.herd_dst ~= nil then
        for k, v in pairs(inst.components.herd_dst.members) do
            k:PushEvent("entermood")
        end
    end
end

local function LeaveMood(inst)
    if inst.components.periodicspawner ~= nil then
        inst.components.periodicspawner:Stop()
    end
    if inst.components.herd_dst ~= nil then
        for k, v in pairs(inst.components.herd_dst.members) do
            k:PushEvent("leavemood")
        end
    end
    inst.components.mood_dst:CheckForMoodChange()
end

local function AddMember(inst, member)
    if inst.components.mood_dst ~= nil then
        member:PushEvent(inst.components.mood_dst:IsInMood() and "entermood" or "leavemood")
    end
end

local function SpawnableParent(inst)
    for member,_ in pairs(inst.components.herd_dst.members) do
        if member.components.domesticatable == nil
            or (member.components.domesticatable:IsDomesticated() == false and member.components.rideable:GetRider() == nil)
            then

            return member
        end
    end
    return nil
end

local function CanSpawn(inst)
    -- Note that there are other conditions inside periodic spawner governing this as well.

    if inst.components.herd_dst == nil or inst.components.herd_dst:IsFull() then
        return false
    end

    local found = SpawnableParent(inst)

    local x, y, z = inst.Transform:GetWorldPosition()
    return found ~= nil
        and #TheSim:FindEntities(x, y, z, inst.components.herd_dst.gatherrange, { "herdmember", inst.components.herd_dst.membertag }) < TUNING.BEEFALOHERD_MAX_IN_RANGE
end

local function OnSpawned(inst, newent)
    --print("At ONSPAWNED",inst)
    if inst.components.herd_dst ~= nil then
        inst.components.herd_dst:AddMember(newent)
    end
    local parent = SpawnableParent(inst)
    if parent ~= nil then
        newent.Transform:SetPosition(parent.Transform:GetWorldPosition())
    end
end

--local function OnFull(inst)
    --TODO: mark some beefalo for death
--end

local function OnInit(inst)
    inst.components.mood_dst:ValidateMood()
end

local function fn()
    local inst = CreateEntity()

    inst.entity:AddTransform()

    inst:AddTag("herd")

    inst:AddTag("NOBLOCK")
    inst:AddTag("NOCLICK")

    inst:AddComponent("herd_dst")
    inst.components.herd_dst:SetMemberTag("beefalo")
    inst.components.herd_dst:SetGatherRange(TUNING.BEEFALOHERD_RANGE)
    inst.components.herd_dst:SetUpdateRange(20)
    inst.components.herd_dst:SetOnEmptyFn(inst.Remove)
    inst.components.herd_dst:SetAddMemberFn(AddMember)

    inst:AddComponent("mood_dst")
    inst.components.mood_dst:SetMoodTimeInDays(TUNING.BEEFALO_MATING_SEASON_LENGTH, TUNING.BEEFALO_MATING_SEASON_WAIT)
    inst.components.mood_dst:SetMoodSeason(SEASONS.SPRING)
    inst.components.mood_dst:SetInMoodFn(InMood)
    inst.components.mood_dst:SetLeaveMoodFn(LeaveMood)
    inst.components.mood_dst:CheckForMoodChange()
    inst:DoTaskInTime(0, OnInit)

    inst:AddComponent("periodicspawner")
    inst.components.periodicspawner:SetRandomTimes(TUNING.BEEFALO_MATING_SEASON_BABYDELAY, TUNING.BEEFALO_MATING_SEASON_BABYDELAY_VARIANCE)
    inst.components.periodicspawner:SetPrefab("babybeefalo")
    inst.components.periodicspawner:SetOnSpawnFn(OnSpawned)
    inst.components.periodicspawner:SetSpawnTestFn(CanSpawn)
    inst.components.periodicspawner:SetDensityInRange(20, 6)
    inst.components.periodicspawner:SetOnlySpawnOffscreen(true)

    return inst
end

return Prefab("beefaloherd_dst", fn, nil, prefabs)
